/*
 * sd.c
 *
 *  Created on: 27 Aug 2011
 *      Author: Mike
 */

#include <sd.h>

uint8 sdCommand(uint8 usartID, uint8 command, uint32 parameter, uint8 response, uint8 tries, uint8 timeout)
{
	uint8 reply, i;
	uint32 loopStartTime, loopTime;
	uint32 startTime;
	uint32 elapsedTime;
	uint8 buffer[6];

	// send command, parameters
	// get reply
	// return 1 if expected reply received
	// return 0 on error
	//====================================
	startTime = getMs();

	debug("sdDoCommand: %2X, parameter: %ld, expected: %02X, tries: %d, timeout: %d secs\n",
			command, parameter, response, tries, timeout);
	reply = 0;

	buffer[0] = command | COMMAND_0;
	buffer[1] = (parameter>>24) & 0xFF;
	buffer[2] = (parameter>>16) & 0xFF;
	buffer[3] = (parameter>> 8) & 0xFF;
	buffer[4] = (parameter>> 0) & 0xFF;
	if (buffer[0]==COMMAND_0)
		buffer[5] = 0x95;
	else
		buffer[5] = 0xFF;	// dummy checksum

	// loop waiting for response as reply
	//===================================
	for(i=0; i<tries; i++)
	{
		// send command, parameter
		//========================
		spiWriteData(usartID, buffer, 6);
		// ShowBuffer(buffer,6);

		loopStartTime = getMs();
		while (1)
		{
			// get reply, check for response
			//==============================
			reply = spiReadByte(usartID);
			if (reply==response)
			{
				// success!
				break;
			}
			sdIdle();
			loopTime = getMs() - loopStartTime;
			if (loopTime > (timeout * 1000))
			{
				// too slow
				//=========
				break;
			}
		}
		if (reply==response)
		{
			// success!
			break;
		}
	}
	if (reply != response)
	{
		// we tried too many times
		//========================
		debug("sdDoCommand: expected %d, got %d\n", response, reply);
		return 0;
	}
	elapsedTime = getMs() - startTime;
	debug("sdDoCommand %02X completed in %d mS\n", elapsedTime);
	return 1;
}

void sdIdle(void)
{
	// service stuff we have to do to keep the world spinning
	//=======================================================
	boardService();
	usbComService();
}

void sdInit(uint8 usartID)
{
	uint16 i;
	uint8 reply;
	uint32 startTime, elapsedTime;

	// initialise SD card
	//===================
	startTime = getMs();

	// initialise SPI
	// USART0, alternate1, polarity negative, phase negative, MSB first
	//=================================================================
	debug("sdInit: initialise\n");
	i = spiInit(usartID, spiAlternate1, spiPolarityN, spiPhaseN, spiBitOrderM);
	debug("sdInit: spiInit returned %d\n", i);

	// deselect the device
	//====================
	debug("sdInit: Deselect\n");
	spiDeselect(usartID);

	// perform low-level SD reset
	//===========================
	debug("sdInit: SDreset\n");
	sdReset(usartID, 80);

	// select the device
	//==================
	debug("sdInit: Select\n");
	spiSelect(usartID);

	// send COMMAND_0
	// want 0x00 as reply
	//===================
	debug("sdInit: COMMAND_0\n");
	reply = sdCommand(usartID, COMMAND_0, 0, 1, MAX_TRIES, 5);
	if (reply==0)
	{
		// didn't get good reply
		//======================
		debug("sdInit: timeout waiting for idle\n");
		return;
	}

	// send COMMAND_1
	// want 0x00 as reply
	//===================
	debug("sdInit: READ_OCR\n");
	reply = sdCommand(usartID, COMMAND_1, 0, 0, MAX_TRIES, 5);
	if (reply==0)
	{
		// we tried too many times
		//========================
		debug("sdInit: READ_OCR timed out\n");

		// deselect the device
		//====================
		spiDeselect(usartID);
		return;
	}

	// deselect the device
	//====================
	spiDeselect(usartID);

	elapsedTime = getMs() - startTime;
	debug("sdInit complete in %ld seconds\n\n", elapsedTime / 1000);
}

uint8 sdReadBlock(uint8 usartID, uint32 LBA, uint8* buffer, uint16 size)
{
	uint16 sizeRead;
	uint32 startTime;
	uint32 endTime;
	uint8 reply;
	uint16 i;
	uint8 checksumBuffer[2];

	// read data from block specified
	// return size read
	// return 0 on error
	//===============================

	// buffer MUST be 512 bytes!
	//==========================
	if (size != 512)
	{
		// invalid buffer size
		//====================
		debug("sdReadBlock buffer size %d invalid MUST be 512\n", size);
		return 0;
	}

	// note the time
	//==============
	startTime = getMs();

	// empty the buffer
	//=================
	for (i=0; i<BUFFER_SIZE; i++)
		buffer[i] = 0;

	// convert block to address
	// block is 512 bytes (2^9)
	//===============================
	LBA <<= BLOCK_SIZE_BITS;

	// padding byte for clocks
	//========================
	spiWriteByte(usartID, 0xFF);

	// read block command
	//==================
	reply = sdCommand(usartID, COMMAND_17, LBA, 0xFE, 3, 1);
	if (reply==0)
	{
		// timed out :(
		//=============
		debug("sdReadData: timeout\n");
		return 0;
	}

	// read block of data
	//===================
	sizeRead = spiReadData(usartID, buffer, size);

	// read the checksum
	// ignore it later!
	//==================
	spiReadData(usartID, checksumBuffer, 2);

	endTime = getMs();
	debug("sdReadData: block read in %ld mS\n", endTime - startTime);
	return sizeRead;
}

void sdReset(uint8 usartID, uint32 count)
{
	uint32 i;
	uint8 port, alternate;
	uint8 highBits, clockBit;
	uint32 delay;
	uint8 selSave, dirSave, portSave;

	// force SD card into native mode
	//===============================

	// need to drive the port directly to ensure correct signal gets generated
	// work out which port and the high and low bit values
	// DI high = MOSI high = TX = 1
	// CS high = SS high = CTS = 1
	// CLK = RTS
	//========================================================================

	if (usartID == 0)
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 1;

		// get the port
		//=============
		if (alternate == 0)
		{
			port = 0;
			highBits = 0x18;
			clockBit = 0x20;
		}
		else
		{
			port = 1;
			highBits = 0x44;
			clockBit = 0x08;
		}
	}
	else
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 2;

		// get the port
		//=============
		if (alternate == 0)
		{
			port = 0;
			highBits = 0x14;
			clockBit = 0x08;
		}
		else
		{
			port = 1;
			highBits = 0x50;
			clockBit = 0x20;
		}
	}

	// set DI and CS high
	// drive clock at least 74 pulses at 400 KHz
	// delay gives slower than 400 KHz
	// but still within range 100-400 KHz
	//==========================================

	delay = 1;

	debug("sdReset port %d, clockBit %2X, count %ld\n", port, clockBit, count);
	if (port==0)
	{
		selSave = P0SEL;
		dirSave = P0DIR;
		portSave = P0;
		P0SEL = 0;
		P0DIR = 0xFF;
		// P0 |= highBits;
		// P0 &= ~highBits;
		P0 = 0xFF;
		for (i=0; i<count; i++)
		{
			P0 |= clockBit;
			delayMicroseconds(delay);
			P0 &= ~clockBit;
			delayMicroseconds(delay);
		}
		P0SEL = selSave;
		P0DIR = dirSave;
		P0 = portSave;
	}
	else
	{
		selSave = P1SEL;
		dirSave = P1DIR;
		portSave = P1;
		P1SEL = 0;
		P1DIR = 0xFF;
		// P1 |= highBits;
		// P1 &= ~highBits;
		P1 = 0xFF;
		for (i=0; i<count; i++)
		{
			P1 |= clockBit;
			delayMicroseconds(delay);
			P1 &= ~clockBit;
			delayMicroseconds(delay);
		}
		P1SEL = selSave;
		P1DIR = dirSave;
		P1 = portSave;
	}
}

